/*
 * adconfig.y:
 */

%{

/* system includes */
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <pwd.h>		/* getpwnam(), getpwuid() */
#include <grp.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>

int (*callback)(void);

#include "configfile.h"
#include "adconfig.h"
#include "adwin-macros.h"
struct yy_config config_dev;
int has_mask, has_gateway;

#include <linux/adwin.h>
adwin_dev_t add_dev;
int del_dev;
int has_ip;




void yyerror(const char *s);
int yylex(void);

#define YYERROR_VERBOSE			1

%}

%union {
	int val;
	char str[40];
}

/*
 * token associate operators
 * <...>: see union above
 */
%token <str> STR
%token <str> IPSTR
%token <str> MACSTR
%token <val> VAL
%token <val> DEVTYPE
%token <val> HELP
%token <str> FLEXERR

/* commands add, del */
%left  <val> UID	/* optional, default "root" */
%left  <val> GID	/* optional, default "adwin" (if exists) else "root" */
%left  <val> MODE	/* optional, default 0666 */
%left  <str> DESC	/* optional, device description text */
%left  <val> TYPE	/* required, one of { link, net, usb } */
%left  <val> BASE	/* required if TYPE=link */
%left  <val> SERIAL	/* required if TYPE=usb */
%left  <str> IP		/* required if TYPE=net */
%left  <val> PORT	/* optional if TYPE=net, default 6543 */
%left  <str> PW		/* optional if TYPE=net, default "" */
%left  <val> HOSTLINK	/* optional if TYPE=net, default 0 */
%left  <val> TIMEOUT	/* optional if TYPE=net, default 1000 */
%left  <val> COUNT	/* optional if TYPE=net, default 5 */

/* command CONFIG */
%left <str> MAC		/* required, MAC to configure */
%left <str> MASK	/* required, subnet mask */
%left <str> GATEWAY	/* optional, gateway, default 0.0.0.0 (no gateway) */
%left <val> DHCP	/* optional, use dhcp, default "false" (0) */

/* list of commands */
%left  <val> ADD
%left  <val> DEL
%left  <val> CONFIG

%type <val> val
%type <val> devtype
%type <str> str
%type <str> ip
%type <str> hwaddr
%type <str> help
%type <str> flexerr

%%
/*
 * Grammar Rules
 */

input:	/* empty */		/* either empty line */
	| input line		/* OR ... ? */
;

line:	  HELP help	{ /* will never reach this (returned) */ }
	| HELP flexerr	{ /* will never reach this (aborted) */ }

line:	  UID val	{ if (! getpwuid(yylval.val)) {
				fprintf(stderr, "warning: no such user %d\n", yylval.val);
			  }
			  add_dev.dev_uid = yylval.val;
			}
	| UID str	{ if (! getpwnam(yylval.str)) {
				fprintf(stderr, "error: no such user %s\n", yylval.str);
				YYABORT;
			  } else {
			  	add_dev.dev_uid = getpwnam(yylval.str)->pw_uid;
			  }
			}
	| UID flexerr	{ /* will never reach this (aborted) */ }

line:	  GID val	{ if (! getgrgid(yylval.val)) {
				fprintf(stderr, "warning: no such group %d\n", yylval.val);
			  } 
			  add_dev.dev_gid = yylval.val;
			}
	| GID str	{ if (! getgrnam(yylval.str)) {
				fprintf(stderr, "error: no such group %s\n", yylval.str);
				YYABORT;
			  }
			  add_dev.dev_gid = getgrnam(yylval.str)->gr_gid;
			}
	| GID flexerr	{ /* will never reach this (aborted) */ }


line:	  MODE val	{ add_dev.dev_mode = yylval.val; }
	| MODE flexerr	{  }

	| TYPE devtype	{ add_dev.dev_type = yylval.val; }
	| TYPE flexerr	{  }

	| PORT val	{ add_dev.dev_net_server.sin_port = htons(yylval.val);
			  config_dev.port = htolel(yylval.val);
			}
	| PORT flexerr	{  }

	| BASE val	{ add_dev.dev_link_base_addr = yylval.val; }
	| BASE flexerr	{  }

	| SERIAL val	{ add_dev.dev_usb_serial = yylval.val; }
	| SERIAL flexerr	{  }

	| HOSTLINK val	{ add_dev.dev_net_host_link = yylval.val; }
	| HOSTLINK flexerr	{  }

	| COUNT val	{ add_dev.dev_net_package_count = yylval.val; }
	| COUNT flexerr	{  }

	| TIMEOUT val	{ add_dev.dev_net_timeout = yylval.val; }
	| TIMEOUT flexerr	{  }

	| DESC str	{ strncpy(add_dev.dev_desc, yylval.str, DESCLEN);
	                  strncpy(config_dev.desc, yylval.str, ETHERNET_DESCLEN);
                        }
	| DESC flexerr	{  }

	| PW str	{ strncpy(add_dev.dev_net_host_password, yylval.str, PWLEN);
			  strncpy(config_dev.pw, yylval.str, PWLEN);
			}
	| PW flexerr	{  }

	| IP ip		{ inet_aton(yylval.str, &add_dev.dev_net_server.sin_addr);
			  inet_aton(yylval.str, (struct in_addr *) config_dev.ip);
			  int32_t *p = (void *)config_dev.ip;
			  *p = __bswap_32(*p);
			  has_ip = 1;
			}
	| IP flexerr	{  }

	| MASK ip	{ inet_aton(yylval.str, (struct in_addr *) config_dev.mask);
			  int32_t *p = (void *)config_dev.mask;
			  *p = __bswap_32(*p);
			  has_mask = 1;
			}
	| MASK flexerr	{  }

	| GATEWAY ip	{ inet_aton(yylval.str, (struct in_addr *) config_dev.gateway);
			  int32_t *p = (void *)config_dev.gateway;
			  *p = __bswap_32(*p);
			  has_gateway = 1;
			}
	| GATEWAY flexerr	{  }

	| DHCP val	{ config_dev.dhcp = htolel(yylval.val); }
	| DHCP flexerr	{  }

	| ADD val	{ callback = add;
	                  clearDevice(&add_dev);
			  add_dev.dev_name = yylval.val;
			  has_ip = 0;	     /* test if type == net */
			}
	| DEL val	{ callback = del;
			  del_dev = yylval.val;
			}
	| CONFIG hwaddr	{ int i;
			  char *ptr;
			  callback = config;
			  config_dev.cmd = htolel(100);
			  config_dev.structVersion = htolel(100);
			  config_dev.dhcp = 0;
			  config_dev.port = htolel(6543);
			  memset(config_dev.pw, '\0', 10);

			  ptr = yylval.str;
			  for (i = 5; i >= 0; i--) {
				config_dev.hwaddr[i] = strtol(ptr, NULL, 16);
				ptr += 3;
			  }
			}
;

help:	  HELP		{ YYABORT; /* YYACCEPT will return 0 .. we don't have a callback ... */ }	;
flexerr:  FLEXERR	{ YYABORT; }			;
devtype:  DEVTYPE	{  }	;
ip:	  IPSTR		{  }	;
hwaddr:	  MACSTR	{  }	;
str:	  STR		{  }	;
val:	  VAL		{  }	;


%%
/*
 * Additional C-Code
 */
#define VALDUMP(x)	printf("%s = %d\n", #x, x)
#define STRDUMP(x)	printf("%s = %s\n", #x, x)



#include <stdio.h>
void yyerror(const char *s) {
//	fprintf(stderr, "%s\n", s);
}








