#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef ADWIN_MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/kernel.h>   /* we want to live in kernel */
#include <linux/module.h>   /* we are a module */
#include <linux/types.h>    /* system independent types __u32 ... */
#include <linux/ioport.h>   /* inb(), out() */

#include <asm/uaccess.h>    /* access user access */

#include <linux/adwin.h>

#ifdef ADWIN_DEBUG
extern int debug;
#define DEBUG(n, args...) if (debug > (n)) printk(args)
#else
#define DEBUG(n, args...)
#endif


/* prototypes of functions defined in this file */
static int register_adwin_resource(int base_addr);
int release_adwin_resource(int base_addr);


static int proc_config_read(struct file * filp, char * buf,
                            size_t nbytes, loff_t *ppos);
static int proc_config_write(struct file *file, const char *buffer,
                             size_t count, loff_t *ppos);
static int proc_config_ioctl(struct inode *inode, struct file *filp,
                             unsigned int cmd, unsigned long arg);


struct file_operations proc_config_fops = {
ioctl:
    proc_config_ioctl,
read:
    proc_config_read,
write:
    proc_config_write,
};

/* This variable is defined and intialized in adwin_main.c
 * We use it to have easy access to the "/proc/adwin/dev" directory when
 * creating/deleting devices.
 */
extern struct proc_dir_entry *dev_root;

/* Keep track on the requested resources.
 */
typedef struct {
    struct list_head list;
    struct resource *res;
    atomic_t count;
} adwin_resource;

LIST_HEAD(adwin_res_list);


/* each device will need a unique minor address. here we keep track of
 * the already used ones. Each newly created device will get a new number.
 * this implies, if you add a device, delete it, add, delete ... you will
 * reach the limit of adwin_minor==255 . Remove the module from the kernel
 * an insert it again to start back from 1.
 */
static atomic_t adwin_minor = ATOMIC_INIT(0);


/*****************************************************************************/
/* helper funtion, map key to corresponding string
 */
char *processor_str (int dp)
{
    switch (dp) {
        case DP_T2:
            return "T2";
        case DP_T4:
            return "T4";
        case DP_T8:
            return "T8";
        case DP_DSP:
            return "DSP";
        case DP_21160:
            return "DSP2";
        default:
            return "N/A";
    }
}

/*****************************************************************************/
/* helper funtion, map key to corresponding string
 */
static char *connection_str (int dc)
{
    switch (dc) {
        case DC_LINK:
            return "link";  /* Link */
        case DC_PCMCIA:
            return "cs";    /* CardService/PCMCIA */
        case DC_USB:
            return "usb";
        case DC_ETHERNET:
            return "net";
        default:
            return "N/A";
    }
}

/*****************************************************************************/
static int register_adwin_resource(int base_addr)
{
    struct list_head *ptr;
    struct resource *res;
    adwin_resource *entry;

    /* request the needed I/O ressources from kernel
     */
    res = request_region(base_addr, 4, MODULE_NAME);
    if (res) {
        /* ok, we are the first ... */
        entry = (adwin_resource *)kmalloc(sizeof(adwin_resource), GFP_KERNEL);
        entry->res = res;

        /* ATOMIC_INIT didn't work, so I used the definition of atomic_t
         */
        entry->count.counter = 1;

        list_add_tail(&entry->list, &adwin_res_list);
        return 0;
    }

    /* ressource is not available, did we already register this addr? */
    for (ptr = adwin_res_list.next; ptr != &adwin_res_list; ptr = ptr->next) {
        entry = list_entry( ptr, adwin_resource, list);

        if (entry->res->start == base_addr) {
            atomic_inc(&entry->count);
            printk(KERN_NOTICE MODULE_NAME ": now %d devices using I/O port %04x:%04x\n",
                   atomic_read(&entry->count), base_addr, base_addr + 3);
            return 0;
        }
    }

    /* someone else has grabbed it, tell user */
    printk(KERN_ERR MODULE_NAME ": failed to register I/O port %04x:%04x\n",
           base_addr, base_addr + 3);
    return -1;
}

/*****************************************************************************/
int release_adwin_resource(int base_addr)
{
    struct list_head *ptr;
    adwin_resource *entry;

    /* free the earlier registered I/ O devices */
    for (ptr = adwin_res_list.next; ptr != &adwin_res_list; ptr = ptr->next) {
        entry = list_entry(ptr, adwin_resource, list);

        if (entry->res->start == base_addr) {
            if (atomic_dec_and_test(&entry->count)) {
                DEBUG(1, "ADwin debug: release %4x\n", base_addr);

                list_del(ptr);
                kfree(entry);
                release_region(base_addr, 4);
            } else
                DEBUG(1, "ADwin debug: %d more devices"
                      " are using register %04x\n",
                      atomic_read(&entry->count), base_addr);
            return 0;
        }
        DEBUG(1, "ADwin debug: %04x ignored, looking for (%04x)\n", entry->res->start, base_addr);
    }

    return -1;
}



/*****************************************************************************/
int proc_dev_create (struct proc_dir_entry *parent, adwin_kdev_t *kdev)
{
    char id[10];
    struct proc_dir_entry *ent;
    sprintf(id, "0x%X", kdev->kdev_name);

    /* I don't know why, but it IS possible to create multiple entries
     * with identical name. We must avoid this, so ...
     */
    DEBUG(1, "ADwin debug: %s already exists ... ", id);
    for (ent = parent->subdir; ent; ent = ent->next)
        if (! strcmp(id, ent->name)) {
            DEBUG(1, "yes\n");
            return -EEXIST;
        }
    DEBUG(1, "no\n");

    if ((kdev->kdev_type == DC_LINK) ||
            (kdev->kdev_type == DC_PCMCIA)) {
        DEBUG(1, "ADwin debug: request_region %04x:%04x ... ",
              kdev->kdev_link_base_addr,
              kdev->kdev_link_base_addr + 3);

        if (register_adwin_resource(kdev->kdev_link_base_addr) < 0) {
            DEBUG(1, "failed (lower)\n");
            return -EINVAL;
        }

        if (register_adwin_resource(kdev->kdev_link_base_addr + 0x10) < 0) {
            DEBUG(1, "failed (upper)\n");
            return -EINVAL;
        }

        DEBUG(1, "ok\n");
    }

    DEBUG(1, "ADwin debug: creating device %s ... ", id);

    ent = proc_mknod(id,
                     S_IFCHR | kdev->kdev_mode,
                     parent,
                     MKDEV(ADWIN_MAJOR, atomic_read(&adwin_minor)));

    if (!ent) {
        DEBUG(1, "failed\n");
        kfree(kdev);
        return -EBADF;
    }
    DEBUG(1, "done\n");

    /* increase the minor, so next device will a unique one
     * TODO: what about a user having more than 255 devices?
     * This should be checked here!
     * (Why I don't do it instead of writing this? I'm too lazy now.)
     */
    atomic_inc(&adwin_minor);

    /* init internal variables */
    atomic_set(&kdev->kdev_in_use, 0);
    atomic_set(&kdev->kdev_use_count, 0);
    init_waitqueue_head(&kdev->kdev_queue);

    /* set up the proc_dir_entry */
    ent->data  = kdev;
    ent->uid   = kdev->kdev_uid;
    ent->gid   = kdev->kdev_gid;
    ent->owner = THIS_MODULE;

    return 0;
}


/*****************************************************************************/
int proc_devices_read(char *page, char **start, off_t off,
                      int count, int *eof, void *data)
{
    char *out = page;
    adwin_kdev_t *kdev;
    struct proc_dir_entry *ptr;

    long ip;
    unsigned char *ips;
    ips = (unsigned char *)&ip;

    for (ptr = dev_root->subdir; ptr; ptr = ptr->next) {
        kdev = (adwin_kdev_t *) (ptr->data);

        out += sprintf(out, "DeviceNo=0x%X ", kdev->kdev_name);
        out += sprintf(out, "Type=%s ", connection_str(kdev->kdev_type));

        switch (kdev->kdev_type) {
            case DC_LINK:
            case DC_PCMCIA:
                out += sprintf(out, "Base=0x%x ",
                               kdev->kdev_link_base_addr);
                break;
            case DC_USB:
                out += sprintf(out, "Serial=%d ",
                               kdev->kdev_usb_serial);
                break;
            case DC_ETHERNET:
                ip = *((long*)&kdev->kdev_net_server.sin_addr);
                out += sprintf(out, "IP=%d.%d.%d.%d ",
                               ips[0], ips[1], ips[2], ips[3]);
                out += sprintf(out, "Port=%d ",
                               ntohs(kdev->kdev_net_server.sin_port));
                out += sprintf(out, "Timeout=%d ", kdev->kdev_net_timeout);
                out += sprintf(out, "PackageCount=%d ", kdev->kdev_net_package_count);
                break;
        }

        /*      out += sprintf(out, "\n       ");
                out += sprintf(out, "%5d", atomic_read(&kdev->kdev_in_use));
                out += sprintf(out, " %d", atomic_read(&kdev->kdev_use_count));
        */      out += sprintf(out, "\n");

    }

    DEBUG(3, "done\n");

    *eof = 1;
    return out - page;
}

/*****************************************************************************/
/* read the configuration file ealier written by the function below
 */

static int proc_config_read(struct file * filp, char * buf,
                            size_t nbytes, loff_t *ppos)
{
    int len, outb, n;
    size_t count;
    int sizeof_dev = sizeof(adwin_dev_t);
    char *page, *out;
    unsigned char *ptr;
    struct proc_dir_entry *ent;
    if (!(page = (char*) __get_free_page(GFP_KERNEL)))
        return -ENOMEM;

    len   = outb = 0;
    count = nbytes;
    for (ent = dev_root->subdir; ent ; ent = ent->next) {
        adwin_dev_t *dev;

        out = page;
        dev = &((adwin_kdev_t *)ent->data)->dev;
        ptr = (char *)dev;
        while ((ptr - (unsigned char *)dev) < sizeof_dev)
            out += sprintf(out, "%02X", *ptr++);

        out += sprintf(out, "\n");
        n = ((out - page) > count) ? count : (out - page);
        if ((len += n) > *ppos) {
            /*
             * self-made modulo-division - a%b gets:
             *  adwin.o: unresolved symbol __moddi3
             *  => ?????
             */
            int offset = *ppos - (int) (((*ppos) /
                                         ((double)sizeof_dev + 1.0)) * (sizeof_dev + 1));
            n = len - *ppos;
            copy_to_user(buf, page + offset, n);
            buf   += n;
            count -= n;
            *ppos += n;  // == len
            outb  += n;
        }
    }
    free_page((unsigned long) page);
    return outb;
}

/*****************************************************************************/
/* write the configuration output
 * If you don't see at once what this code is doing `cat /proc/adwin/config`
 * and you will know :)
 */
static int proc_config_write(struct file *file, const char *buffer,
                             size_t count, loff_t *ppos)
{

    const int sizeof_dev = sizeof(adwin_dev_t);
    u_char data[2 * sizeof_dev + 1] = { 0 };
    u_char *data_ptr, *ptr = (u_char *) buffer;
    adwin_kdev_t *kdev;

    u_char *kdev_ptr;

    while (count >= (2 * sizeof_dev + 1)) {
        copy_from_user(data, ptr, 2 * sizeof_dev + 1);
        data_ptr = data;

        kdev = (adwin_kdev_t *) kmalloc(sizeof(adwin_kdev_t), GFP_KERNEL);
        kdev_ptr = (u_char *) &kdev->dev;

        while ((kdev_ptr - (u_char *)kdev) < sizeof_dev) {
            if ((*data_ptr >= 0x30) && (*data_ptr <= 0x39))
                *kdev_ptr = (*data_ptr - 0x30) << 4;
            else if ((*data_ptr >= 0x41) && (*data_ptr <= 0x46))
                *kdev_ptr = (*data_ptr - 0x37) << 4;
            data_ptr++;

            if ((*data_ptr >= 0x30) && (*data_ptr <= 0x39))
                *kdev_ptr |= *data_ptr - 0x30;
            else if ((*data_ptr >= 0x41) && (*data_ptr <= 0x46))
                *kdev_ptr |= *data_ptr - 0x37;
            data_ptr++;

            kdev_ptr++;
        }

        //      kdev->kdev_btl_version = 0;
        //      kdev->kdev_card_version = 0;
        kdev->kdev_processor   = DP_UNKNOWN;

        proc_dev_create(dev_root, kdev);

        ptr   += 2 * sizeof_dev + 1;
        count -= 2 * sizeof_dev + 1;
    }

    return ptr - (unsigned char *)buffer;
}

/*****************************************************************************/
static int proc_config_ioctl(struct inode *inode, struct file *filp,
                             unsigned int cmd, unsigned long arg)
{

    char id[10];
    adwin_kdev_t *kdev;
    struct proc_dir_entry *p;
    int dev_size  = sizeof(adwin_dev_t);
    int kdev_size = sizeof(adwin_kdev_t);

    if (! capable(CAP_SYS_ADMIN))
        return -EPERM;

    switch (cmd) {
        case ADIOKDEVADD:
            kdev = (adwin_kdev_t *)kmalloc(kdev_size, GFP_KERNEL);
            copy_from_user(&kdev->dev, (adwin_dev_t *) arg, dev_size);
            return proc_dev_create(dev_root, kdev);

        case ADIOKDEVDEL:
            sprintf(id, "0x%X", (int)arg);
            DEBUG(1, "ADwin debug: removing %s", id);
            for (p = dev_root->subdir; p; p = p->next)
                if (! strcmp(p->name, id)) {
                    adwin_kdev_t *kdev = (adwin_kdev_t *)p->data;
                    if (kdev->kdev_type == DC_LINK) {
                        DEBUG(1, " (ioports)");
                        release_adwin_resource(kdev->kdev_link_base_addr);
                        release_adwin_resource(kdev->kdev_link_base_addr + 0x10);
                    }
                    DEBUG(1, " (memory)");
                    kfree(kdev);
                    DEBUG(1, " (proc)");
                    remove_proc_entry(p->name, dev_root);
                    DEBUG(1, "... done\n");
                    return 0;
                }
            DEBUG(1, " ... not found\n");
            return -ENODEV;

        default:
            return -ENOTTY;
    }
}
